<?php

namespace Drupal\deprecation_status;

use Drupal\Core\File\FileSystemInterface;
use GuzzleHttp\Exception\RequestException;

class DataSource {

  /**
   * List of file names for all dataset files.
   *
   * @type array
   */
  public static $datafiles = [
    'metadata.csv',
    'core_deprecation_count.csv',
    'breakdown_top200_error_counts.csv',
    'breakdown_top200_error_types.csv',
    'breakdown_top200_projects.csv',
    'breakdown_total_error_counts.csv',
    'breakdown_total_error_types.csv',
    'breakdown_total_projects.csv',
    'result_errors.csv',
    'result_projects.csv',
  ];

  /**
   * Get full path of data file based on filename.
   *
   * @param string $filename
   *   Filename. For example 'metadata.csv'.
   * @return string
   *   Full path to $filename.
   */
  public static function getFullPath($filename) {
    // Attempt to use an up to date copy of the data from locally saved files.
    /* @var \Drupal\file\FileInterface[] $files */
    $files = \Drupal::entityTypeManager()
      ->getStorage('file')
      ->loadByProperties(['uri' => 'private://deprecation_status_' . $filename]);
    if (count($files)) {
      $file = array_shift($files);
      return $file->getFileUri();
    }

    // If that fails, we still have shipped files locally.
    else {
      return drupal_get_path('module', 'deprecation_status') . '/data/' . $filename;
    }
  }

  /**
   * Get metadata about current dataset.
   *
   * @return array
   *   Two element indexed array containing the dataset ID and the dataset date.
   */
  public static function getMetaData() {
    $metadata = file_get_contents(static::getFullPath('metadata.csv'));
    return explode(';', trim($metadata));
  }

  /**
   * Get metadata about shipped dataset.
   *
   * As opposed to getMetaData() this will always use the local shipped file.
   *
   * @return array
   *   Two element indexed array containing the dataset ID and the dataset date.
   */
  public static function getShippedMetaData() {
    $path = drupal_get_path('module', 'deprecation_status') . '/data/metadata.csv';
    $metadata = file_get_contents($path);
    return explode(';', trim($metadata));
  }

  /**
   * Download and update all local dataset files.
   *
   * @return string|boolean
   *   TRUE if all successful. Error string otherwise.
   */
  public static function updateDataFiles() {
    $datastore = [];

    $client = \Drupal::httpClient();
    try {
      // Try downloading each dataset file. Store them in memory as a protection
      // against half updates. Only save locally once all files downloaded
      // successfully.
      foreach(static::$datafiles as $filename) {
        $response = $client->get('https://git.drupalcode.org/project/deprecation_status/raw/8.x-2.x/data/' . $filename . '?' . time());
        $data = (string) $response->getBody();
        if (!empty($data)) {
          $datastore[$filename] = $data;
        }
        else {
          return 'Missing data in ' . $filename . '.';
        }
      }
    }
    catch (RequestException $e) {
      return 'HTTP error while downloading online dataset.';
    }

    // Save updated files locally as private files.
    foreach($datastore as $filename => $data) {
      if (!file_save_data($data, 'private://deprecation_status_' . $filename, FileSystemInterface::EXISTS_REPLACE)) {
        return 'Error while saving updated ' . $filename . ' locally.';
      }
    }

    return TRUE;
  }

  /**
   * Download and update all local dataset files.
   *
   * @return string|boolean
   *   TRUE if all successful. Error string otherwise.
   */
  public static function resetDataFiles() {
    $count = 0;
    foreach(static::$datafiles as $filename) {
      /* @var \Drupal\file\FileInterface[] $files */
      $files = \Drupal::entityTypeManager()
        ->getStorage('file')
        ->loadByProperties(['uri' => 'private://deprecation_status_' . $filename]);
      foreach($files as $file) {
        $file->delete();
        $count++;
      }
    }
    return $count;
  }

}
