<?php

namespace Drupal\deprecation_status\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\deprecation_status\DataSource;
use Symfony\Component\HttpFoundation\RedirectResponse;

class ProjectsForm extends FormBase {

  /**
   * Stored sorting criteria.
   *
   * @type int
   */
  protected $column;

  /**
   * Stored sorting criteria ('asc' or 'desc').
   *
   * @type string
   */
  protected $sort;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'deprecation_status_projects_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['#cache']['contexts'][] = 'session';
    $form['#attached']['library'][] = 'deprecation_status/deprecation_status.lists';

    $form['filters'] = [
      '#type' => 'details',
      '#title' => $this->t('Filter projects'),
      '#attributes' => ['class' => ['container-inline']],
      '#open' => TRUE,
    ];

    // Backwards compatibility for the old 'name' filter.
    if (!empty($_GET['name']) && empty($_GET['names'])) {
      $_GET['names'] = '*' . $_GET['name'] . '*';
      unset($_GET['name']);
      return new RedirectResponse(Url::fromRoute('deprecation_status.projects_form', [], ['query' => $_GET])->toString());
    }

    $form['filters']['names'] = [
      '#title' => 'Names',
      '#type' => 'textfield',
      '#default_value' => @$_GET['names'],
      '#size' => 20,
    ];
    $form['filters']['result'] = [
      '#title' => 'Result',
      '#type' => 'select',
      '#options' => [
        'Explicitly Drupal 9 compatible' => 'Explicitly Drupal 9 compatible',
        'Not explicitly Drupal 9 compatible' => 'Not explicitly Drupal 9 compatible',
        'No problems found, info.yml change needed' => '-- No problems found, info.yml change needed',
        'All rector covered' => '-- All rector covered',
        'Some covered by rector' => '-- Some covered by rector',
        'Needs manual review' => '-- Needs manual review',
        'No results' => '-- No results',
        'Invalid results' => '-- Invalid results',
      ],
      '#default_value' => @$_GET['result'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['error'] = [
      '#title' => 'Error',
      '#type' => 'select',
      '#options' => [
        'Environment' => 'Environment',
        'Composer' => 'Composer',
      ],
      '#default_value' => @$_GET['error'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['type'] = [
      '#title' => 'Type',
      '#type' => 'select',
      '#options' => [
        'Module' => 'Module',
        'Theme' => 'Theme',
        'Theme_engine' => 'Theme engine',
        'Distribution' => 'Distribution',
        'Core' => 'Core',
      ],
      '#default_value' => @$_GET['type'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['group'] = [
      '#title' => 'Group',
      '#type' => 'select',
      '#options' => [],
      '#default_value' => @$_GET['group'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    foreach([50, 100, 200, 500, 1000] as $number) {
      $form['filters']['group']['#options'][$number] = '<= ' . $number;
    }
    $form['filters']['plan'] = [
      '#title' => 'Drupal 9 plan',
      '#type' => 'select',
      '#options' => ['none' => 'None', 'specified' => 'Specified'],
      '#default_value' => @$_GET['plan'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['patch'] = [
      '#title' => 'Rector patch',
      '#type' => 'select',
      '#options' => ['none' => 'None', 'available' => 'Available'],
      '#default_value' => @$_GET['patch'],
      '#required' => FALSE,
      '#empty_option' => '- No filter -',
    ];
    $form['filters']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    // Read all projects into one big array.
    $projects = [];
    $file = fopen(DataSource::getFullPath('result_projects.csv'), 'r');
    $i = 0;
    while ($line = fgetcsv($file, 0, ";")) {
      // Skip the header row.
      if ($i > 0) {
        $patch = array_pop($line);
        array_pop($line);
        array_pop($line);
        $line[] = $patch;
        $projects[] = $line;
      }
      $i++;
    }
    fclose($file);

    // Set up columns for the table.
    $columns = [
      'name'    => 'Name',
      'result'  => 'Result',
      'error'   => 'Error',
      'type'    => 'Project type',
      'usage'   => 'Usage',
      'group'   => 'Group',
      'total'   => 'Total deprecations',
      'parsing' => 'Total parsing errors',
      'rector'  => 'Covered by rector',
      'drupal'  => 'Not covered by rector',
      'na'      => 'Non-Drupal deprecations',
      'plan'    => 'Drupal 9 plan',
      'patch'  => 'Rector patch',
    ];
    // Add sorting information to columns.
    $i = 0;
    foreach ($columns as $name => &$column) {
      $column = ['data' => $column, 'field' => $i];
      $i++;
    }
    // Sort by usage descending initially.
    $columns['usage']['sort'] = 'desc';

    // Sort the data array based on query tablesort information.
    $order = tablesort_get_order($columns);
    $this->column = $order['sql'];
    $this->sort = tablesort_get_sort($columns);
    uasort($projects, [$this, 'tableSort']);

    $common_script = <<<SCRIPTEND
    responsive: false,
    legend: {
      display: true,
      position: window.innerWidth < 800 ? 'top' : 'right'
    },
    tooltips: {
      callbacks: {
        label: function(tooltipItem, data) {
          var dataset = data.datasets[tooltipItem.datasetIndex];
          var meta = dataset._meta[Object.keys(dataset._meta)[0]];
          var total = meta.total;
          var currentValue = dataset.data[tooltipItem.index];
          var percentage = parseFloat((currentValue/total*100).toFixed(1));
          return currentValue + ' (' + percentage + '%)';
        },
        title: function(tooltipItem, data) {
          return data.labels[tooltipItem[0].index];
        }
      }
    }
SCRIPTEND;
    $form['chart'] = [
      '#children' => <<<SCRIPTEND
<script src="https://cdn.jsdelivr.net/npm/chart.js@2.8.0"></script>
<div id="charts">
 <canvas id="totalProjects"></canvas>
 <canvas id="totalDeprecations"></canvas>
</div>
<script type="text/javascript">
var ctx = document.getElementById('totalProjects').getContext('2d');
var chart = new Chart(ctx, {
  type: 'doughnut',
  data: {
    labels: ['Explicitly Drupal 9 compatible', 'No problems found, info.yml change needed', 'All rector covered', 'Some covered by rector', 'Needs manual review', 'No results', 'Invalid results'],
    datasets: [{
      data: [@totalProjectsData],
      backgroundColor: ["#8CBD3F", "#FDC70F", "#1F72B2", "#42509B", "#EC6225", "#000000"]
    }]
  },
  options: {
    aspectRatio: window.innerWidth < 800 ? 1.5 : 2,
    title: {
      display: true,
      text: '@projectcount projects by results',
      position: 'bottom'
    },
    $common_script
  }
});
var ctx = document.getElementById('totalDeprecations').getContext('2d');
var chart = new Chart(ctx, {
  type: 'doughnut',
  data: {
    labels: ['Covered by rector', 'Not covered by rector', 'Non-Drupal core deprecations', 'Parsing errors'],
    datasets: [{
      data: [@totalDeprecationsData],
      backgroundColor: ["#1F72B2", "#EC6225", "#C31A7F", "#F1E70D"]
    }]
  },
  options: {
    aspectRatio: window.innerWidth < 800 ? 1.5 : 2,
    title: {
      display: true,
      text: '@deprecatedcount errors found',
      position: 'bottom'
    },
    $common_script
  }
});
</script>
SCRIPTEND
    ];

    // Display the resulting data.
    $form['table'] = [
      '#type' => 'table',
      '#header' => $columns,
      '#empty' => t('No matching projects found.'),
    ];

    // Filter projects down based on form input. Add data summaries of overall
    // results or filtered results.
    $project_count = count($projects);
    $summary = $this->dataSummary($projects);
    $projects = $this->tableFilter($projects);
    if (count($projects) != $project_count) {
      $summary = $this->dataSummary($projects);
    }
    if (count($projects) == 0) {
      unset($form['chart']);
    }
    else {
      $project_data = join(',', array_values($summary['results']));
      $deprecation_data = join(',', array_values($summary['errors']));
      $deprecation_count = $summary['total'];
      $form['chart']['#children'] = str_replace(
        ['@totalProjectsData', '@totalDeprecationsData', '@projectcount', '@deprecatedcount'],
        [$project_data, $deprecation_data, count($projects), $deprecation_count],
        $form['chart']['#children']
      );
    }

    // Page the results based on query information.
    $page = pager_find_page();
    $num_per_page = 30;
    pager_default_initialize(count($projects), $num_per_page);
    $offset = $num_per_page * $page;
    $project_page = array_slice($projects, $offset, $num_per_page);

    // Get dataset metadata for outside links to error logs.
    list ($data_id, $data_date) = DataSource::getMetaData();

    // Key results based on header key names.
    $keys = array_keys($columns);
    foreach ($project_page as $project) {
      $info = [];
      if (in_array($project[1], ['No results', 'Invalid results'])) {
        $log_uri =
          'https://dispatcher.drupalci.org/job/phpstan/' . $data_id .
          '/artifact/phpstan-results/' . str_replace(' ', '.', $project[0]) . '.phpstan_stderr/*view*/';
        $info['name'] = [
          '#type' => 'markup',
          '#markup' => $project[0],
        ];
        $info['result'] = [
          'data' => [
            '#type' => 'link',
            '#title' => $project[1],
            '#url' => Url::fromUri($log_uri),
            '#attributes' => ['class' => ['category']],
          ],
        ];
        $class = 'unknown';
      }
      else {
        $info['name'] = [
          'data' => [
            '#type' => 'link',
            '#title' => $project[0],
            '#url' => Url::fromRoute('deprecation_status.errors_form', [], ['query' => ['project' => $project[0]]])
          ]
        ];
        $info['result'] = [
          '#type' => 'markup',
          '#markup' => '<span class="category">' . $project[1] . '</span>',
        ];
        $class = 'warning';
        if ($project[1] == 'Explicitly Drupal 9 compatible') {
          $class = 'passed';
        }
        elseif (in_array($project[1], ['All rector covered', 'Some covered by rector', 'Needs manual review'])) {
          $class = 'error';
        }
      }
      $info['#attributes']['class'] = ['item-' . $class];

      foreach($project as $index => $item) {
        if (!isset($info[$keys[$index]])) {
          $info[$keys[$index]] = [
            '#type' => 'markup',
            '#markup' => $item,
          ];
        }
      }
      // Format drupal.org issue links in plans and fix relative links.
      $plan = preg_replace('!(\[#([0-9]+)\])!', '<a target="_blank" href="https://drupal.org/node/\2">\1</a>', $project[11]);
      $plan = preg_replace('!href="/!', 'href="https://drupal.org/', $plan);
      $info['plan'] = [
        '#type' => 'markup',
        '#markup' => $plan,
      ];
      $info['total']['#markup'] = '<strong>' . $info['total']['#markup'] . '</strong>';

      if (!empty($project[12]) && $project[1] != 'Explicitly Drupal 9 compatible') {
        $patch_uri = 'https://dispatcher.drupalci.org/job/phpstan/' . $data_id . '/artifact/phpstan-results/' . $project[12] . '/*view*/';
        $info['patch'] = [
          'data' => [
            '#type' => 'link',
            '#title' => $this->t('Available'),
            '#url' => Url::fromUri($patch_uri),
          ],
        ];
      }
      else {
        $info['patch'] = [
          '#type' => 'markup',
          '#markup' => ''
        ];
      }

      $form['table'][] = $info;
    }

    $form['pager'] = array(
      '#type' => 'pager'
    );

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $keys = ['names', 'result', 'error', 'type', 'group', 'plan', 'patch'];
    $submitted = array_filter($form_state->getValues());
    $query = [];
    foreach ($keys as $key) {
      if (!empty($submitted[$key])) {
        $query[$key] = $submitted[$key];
      }
    }
    $form_state->setRedirect('deprecation_status.projects_form', [], ['query' => $query]);
  }

  /**
   * Custom table sorting based on stored sorting criteria.
   */
  protected function tableSort($a, $b) {
    if ($a[$this->column] == $b[$this->column]) {
      return 0;
    }
    $cmp = ($a[$this->column] < $b[$this->column]) ? -1 : 1;
    return ($this->sort == 'desc') ? $cmp * -1 : $cmp;
  }

  /**
   * Custom table filter based on form values.
   */
  protected function tableFilter($projects) {
    $keys = ['names', 'result', 'error', 'type', 'group', 'plan', 'patch'];
    foreach ($keys as $key) {
      $$key = @$_GET[$key];
    }

    // Process peoject name matching for preg_match() later.
    $name_list = [];
    if (!empty($names)) {
      $all_names = explode(',', $names);
      foreach($all_names as $one_name) {
        $parts = explode('*', trim($one_name));
        $processed_parts = [];
        foreach ($parts as $part) {
          $processed_parts[] = preg_quote($part, '!');
        }
        $name_list[] = join('.*', $processed_parts);
      }
    }

    $filtered = [];
    foreach ($projects as $project) {
      // Assume this row is a match.
      $match = TRUE;

      // Go through each filter and if the row is still a match, see if this
      // filter will invalidate it.
      if (!empty($name_list) && $match) {
        $one_match = FALSE;
        foreach($name_list as $name) {
          if (preg_match("!^$name$!", $project[0])) {
            $one_match = TRUE;
            break;
          }
        }
        $match = $one_match;
      }
      if (!empty($result) && $match) {
        if ($result == 'Not explicitly Drupal 9 compatible') {
          $match = ($project[1] != 'Explicitly Drupal 9 compatible');
        }
        else {
          $match = ($project[1] == $result);
        }
      }
      if (!empty($error) && $match) {
        $match = ($project[2] == $error);
      }
      if (!empty($type) && $match) {
        $match = ($project[3] == $type);
      }
      if (!empty($group) && $match) {
        $match = ($project[5] <= $group);
      }
      if (!empty($plan) && $match) {
        $match = ($plan == 'none' ? empty($project[11]) : !empty($project[11]));
      }
      if (!empty($patch) && $match) {
        $match = ($patch == 'none' ? empty($project[12]) : !empty($project[12]) && $project[1] != 'Explicitly Drupal 9 compatible');
      }

      // If the row was still a match, continue.
      if ($match) {
        $filtered[] = $project;
      }
    }

    return $filtered;
  }

  protected function dataSummary($projects) {
    $summary = [
      'projects' => count($projects),
      'total' => 0,
      'errors' => [
        'rector' => 0,
        'drupal' => 0,
        'na' => 0,
        'parsing' => 0,
      ],
      'results' => [
        'Explicitly Drupal 9 compatible' => 0,
        'No problems found, info.yml change needed' => 0,
        'All rector covered' => 0,
        'Some covered by rector' => 0,
        'Needs manual review' => 0,
        'No results' => 0,
        'Invalid results' => 0,
      ],
    ];
    foreach($projects as $project) {
      $summary['results'][$project[1]]++;
      if ($project[1] != 'Explicitly Drupal 9 compatible') {
        $summary['total'] += $project[6] + $project[7];
        $summary['errors']['rector'] += $project[8];
        $summary['errors']['drupal'] += $project[9];
        $summary['errors']['na'] += $project[10];
        $summary['errors']['parsing'] += $project[7];
      }
    }
    return $summary;
  }

}
